(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('tslib')) :
    typeof define === 'function' && define.amd ? define(['exports', 'tslib'], factory) :
    (global = global || self, factory(global.systyle = {}, global.tslib_1));
}(this, function (exports, tslib_1) { 'use strict';

    function isMergeable(value) {
        return value !== null && typeof value === 'object' && value.constructor === Object; // prettier-ignore
    }
    function flatten(arr, res) {
        if (res === void 0) { res = []; }
        var i, cur;
        var len = arr.length;
        for (i = 0; i < len; i++) {
            cur = arr[i];
            Array.isArray(cur) ? flatten(cur, res) : res.push(cur);
        }
        return res;
    }
    function merge(target, source) {
        if (!isMergeable(target) || !isMergeable(source))
            return source;
        var merged = tslib_1.__assign({}, target);
        for (var key in source) {
            merged[key] = merge(target[key], source[key]);
        }
        return merged;
    }
    function apply(acc, fn) {
        return fn(acc) || acc;
    }
    function compose(fns) {
        return function (input) { return fns.reduceRight(apply, input); };
    }
    function moulinettify(declaration) {
        return isMergeable(declaration)
            ? function (props) { return merge(declaration, props); }
            : declaration;
    }

    function createSystem(builder, moulinettes, wrappers, extensions) {
        if (moulinettes === void 0) { moulinettes = []; }
        if (wrappers === void 0) { wrappers = []; }
        if (extensions === void 0) { extensions = []; }
        var moulinette = compose(moulinettes);
        var wrap = compose(wrappers);
        var System = wrap(builder(moulinette));
        System.compute = moulinette;
        System.with = function () {
            var extras = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                extras[_i] = arguments[_i];
            }
            return createSystem(builder, moulinettes.concat(flatten(extras).map(moulinettify)), wrappers, extensions);
        }; // prettier-ignore
        System.wrap = function () {
            var extras = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                extras[_i] = arguments[_i];
            }
            return createSystem(builder, moulinettes, wrappers.concat(extras), extensions);
        };
        System.extend = function (extension) {
            return createSystem(builder, moulinettes, wrappers, extensions.concat([extension]));
        };
        extensions.forEach(function (extend) { return extend(System); });
        return System;
    }

    var rxRules = /^(\&|\@|\:|\#|\.)/;
    function isCSS(_, key) {
        return rxRules.test(key) || key in document.body.style;
    }
    function isEmpty(value) {
        return value === null || (typeof value === 'object' && Object.keys(value).length === 0); // prettier-ignore
    }
    function isObject(value) {
        return value !== null && typeof value === 'object' && value.constructor === Object; // prettier-ignore
    }
    function partition(by) {
        return function (props) {
            var match = {};
            var rest = {};
            Object.keys(props).forEach(function (key) {
                var value = props[key];
                if (by(value, key, props)) {
                    match[key] = value;
                }
                else {
                    rest[key] = value;
                }
            });
            return [match, rest];
        };
    }
    function compact(props) {
        return partition(isEmpty)(props)[1];
    }
    function replace(by) {
        var _replace = function (props, context) {
            if (context === void 0) { context = props; }
            var result = {};
            for (var key in props) {
                var parsed = by(props[key], key, context);
                var value = typeof parsed === 'undefined' ? props[key] : parsed;
                result[key] = isObject(value) ? _replace(value, context) : value;
            }
            return result;
        };
        return _replace;
    }
    function rename(aliases) {
        var _rename = function (props) {
            var renamed = {};
            Object.keys(props).forEach(function (key) {
                var value = props[key];
                var alias = key in aliases ? aliases[key] : key;
                renamed[alias] = isObject(value) ? _rename(value) : value;
            });
            return renamed;
        };
        return _rename;
    }

    var aliases = rename({
        bg: 'backgroundColor',
        mt: 'marginTop',
        mr: 'marginRight',
        mb: 'marginBottom',
        ml: 'marginLeft',
        pt: 'paddingTop',
        pr: 'paddingRight',
        pb: 'paddingBottom',
        pl: 'paddingLeft',
        font: 'fontFamily',
        size: 'fontSize'
    });

    function theme(props, category, key) {
        if (!('theme' in props) || !(category in props.theme))
            return key;
        var values = props.theme[category];
        if (typeof key !== 'undefined') {
            return key in values ? values[key] : key;
        }
        else {
            return values;
        }
    }
    var rxSpacing = /margin|padding/;
    var rxColor = /color|Color/;
    var replaceCSS = function (parser) {
        return replace(function (value, key, props) {
            return isCSS(value, key) ? parser(value, key, props) : undefined;
        });
    };
    var spacing = replaceCSS(function (value, key, props) {
        if (typeof value === 'number' && rxSpacing.test(key)) {
            return value * (theme(props, 'spacing') || 1);
        }
    });
    var fonts = replaceCSS(function (value, key, props) {
        if (key === 'fontFamily') {
            return theme(props, 'fonts', value);
        }
    });
    var sizes = replaceCSS(function (value, key, props) {
        if (key === 'fontSize') {
            return theme(props, 'sizes', value);
        }
    });
    var colors = replaceCSS(function (value, key, props) {
        if (key === 'background' || rxColor.test(key)) {
            return theme(props, 'colors', value);
        }
    });
    var applyTheme = compose([spacing, fonts, sizes, colors]);

    var partitionCSS = partition(isCSS);
    function extractCSS(input) {
        var _a = partitionCSS(input), css = _a[0], props = _a[1];
        var definedCSS = compact(css);
        if (!isEmpty(definedCSS)) {
            props.css = props.css || [];
            props.css = [css].concat(props.css);
        }
        return props;
    }
    var compileCSS = compose([extractCSS, applyTheme, aliases]);

    function stringify(strings, args, props) {
        var computed = args.map(function (arg) {
            return typeof arg === 'function' ? arg(props) : arg;
        });
        return strings.reduce(function (template, string, i) {
            return template + string + (i in computed ? computed[i] : '');
        }, '');
    }
    function addTemplate(strings, args) {
        return function (input) {
            var _a = compileCSS(input) || {}, _b = _a.css, css = _b === void 0 ? [] : _b, props = tslib_1.__rest(_a, ["css"]);
            var str = stringify(strings, args, props);
            return tslib_1.__assign({}, props, { css: [str].concat(css) });
        };
    }

    function createElement(_a) {
        var _b = _a.as, type = _b === void 0 ? 'div' : _b, props = tslib_1.__rest(_a, ["as"]);
        return tslib_1.__assign({}, props, { as: type });
    }

    function addAnimation(nextAnimation) {
        return function (_a) {
            var _b = _a.animation, animation = _b === void 0 ? '' : _b, _c = _a.animations, animations = _c === void 0 ? [] : _c, props = tslib_1.__rest(_a, ["animation", "animations"]);
            return (tslib_1.__assign({}, props, { animations: animations.concat([nextAnimation.trim(), animation]).filter(Boolean) }));
        };
    }
    function combineAnimations(_a) {
        var _b = _a.animations, animations = _b === void 0 ? [] : _b, props = tslib_1.__rest(_a, ["animations"]);
        return tslib_1.__assign({}, props, { animation: animations.join() });
    }

    function extension(Styled) {
        Styled.as = function (type) { return Styled.with({ as: type }); };
        Styled.css = function (strings) {
            var args = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                args[_i - 1] = arguments[_i];
            }
            return Styled.with(addTemplate(strings, args));
        };
        Styled.animate = function (duration, animation) {
            if (animation === void 0) { animation = ''; }
            return Styled.with(addAnimation(animation + " " + duration));
        };
    }
    function createStyled(builder) {
        return createSystem(builder)
            .extend(extension)
            .with(createElement, compileCSS, combineAnimations);
    }

    exports.extension = extension;
    exports.createStyled = createStyled;
    exports.default = createStyled;

    Object.defineProperty(exports, '__esModule', { value: true });

}));
